#include <NanoESP.h>
#include <NanoESP_HTTP.h>                                    //special library for http-connection
#include <SoftwareSerial.h>
#include "Websites.h"                                        //websites in a special file as progmem

#define SSID "<<SSID>>"
#define PASSWORD "<<PASSWORD>>"

#define LED_WLAN 13

#define R1  2                                                // Digital I/O pin numbers of the relais
#define R2  3
#define R3  4
#define R4  5
#define R5  6
#define R6  7
#define R7  8
#define R8  14

#define TRIGGER 9
#define ECHO 10


#define MAXMIN (1000UL * 60 * 20)                           // max pump on time (20 min)
unsigned long lRollTime = millis() + MAXMIN;

NanoESP oNanoESP = NanoESP();
NanoESP_HTTP oHttp = NanoESP_HTTP(oNanoESP);                //init http library with an instance of nanoesp

SoftwareSerial oSensor =  SoftwareSerial(9,10); 

bool bIsRelayOn = false;
byte read_buffer[4];
int i, j;
long lDistance = 0;
String sDistance = "";

//------- Start initialize -----------------------------------------------------------------------------------

void setup() {
  Serial.begin(9600);
  Serial.println("start");

  oNanoESP.init();
  
  oNanoESP.sendCom("AT+CWDHCP_CUR=1,0");                    // set static IP
  oNanoESP.sendCom("AT+CIPSTA_CUR=\"192.168.1.145\",\"192.168.1.100\",\"255.255.255.0\"");
  
                                                            //! Only if you use a new network
  oNanoESP.configWifi(DUAL, SSID, PASSWORD);
    
                                                            // Station Mode (receiver only - no access point)
  oNanoESP.configWifiMode(STATION);
  
                                                            // Wait till Connected to WiFi
  if (oNanoESP.wifiConnected())digitalWrite(LED_WLAN, HIGH); else Serial.println(F("Error WLAN Connect"));

                                                            //Start TCP Server
  if (oNanoESP.startTcpServer(80))    Serial.println(F("TCP Server Activ")); else   Serial.println(F("TCP Server Error"));

  Serial.println(oNanoESP.getIp());                         // Print IP to Terminal

  digitalWrite(R1, HIGH);                                   // Initialize Pins so relays are inactive at reset
  digitalWrite(R2, HIGH);
  digitalWrite(R3, HIGH);
  digitalWrite(R4, HIGH);  
  digitalWrite(R5, HIGH);  
  digitalWrite(R6, HIGH);  
  digitalWrite(R7, HIGH);  
  digitalWrite(R8, HIGH);  
  
  pinMode(R1, OUTPUT);                                      // set pins as outputs
  pinMode(R2, OUTPUT);  
  pinMode(R3, OUTPUT);  
  pinMode(R4, OUTPUT);    
  pinMode(R5, OUTPUT);
  pinMode(R6, OUTPUT);
  pinMode(R7, OUTPUT);
  pinMode(R8, OUTPUT);

  pinMode(TRIGGER, OUTPUT);
  pinMode(ECHO, INPUT);
  digitalWrite(TRIGGER, HIGH);                             //Signal abschalten

  delay(2000); //Check that all relays are inactive at Reset
}

//------- Start main loop -----------------------------------------------------------------------------------

void loop() {
  String sMethod, sResource, sParameter, sDevice, sAction;
  char cRelayNo;
  int id, iRelayNo;

  oNanoESP.listen();

  if (oHttp.recvRequest(id, sMethod, sResource, sParameter)) { //Incoming request, parameters by reference

    Serial.println("New Request from id :" + String(id) + ", method: " + sMethod +  ", resource: " + sResource +  " parameter: " + sParameter);

    if (sParameter != "") {
        sDevice = sParameter.substring(1,sParameter.indexOf('='));
        sAction = sParameter.substring(sParameter.indexOf('=') + 1);
    }

    if (sResource == "/") {                                   //index page
      oHttp.sendFromFlash(id, index, sizeof(index));          //send website from website.h
      oNanoESP.closeConnection(id);                           //close connection so browser knows website transmission is complete
    }
    
    else if (sResource == "/device/") {

      if (sDevice == "US1") { // Sensor

        i = 0;
        j = 0;
        sDistance = "";
        while(lDistance == 0 || j < 3){                       // three measures to calculate the average in FHEM
          lDistance = getDistance();
          if (lDistance != 0) {
            if (sDistance != "") sDistance = sDistance + "|";
            sDistance = sDistance + String(lDistance);
            j++;
          }
          i++;
          if (i == 20) break;                                 // emergency stop => exit after 20 loops contains some reserve
          delay(1000);                                        // one second between measures; lower value => more variations
        }
        oNanoESP.sendDataClose(id, sDistance);

      } 
      
      else if (sDevice == "STS") {                            // get the status of all relais
        
         oNanoESP.sendDataClose(id, getRelayStatusAll());
         
      } else {                                                // switch relais and return the status of the current relay
        cRelayNo = sDevice.charAt(1);
        if(isDigit(cRelayNo)) {
          iRelayNo = cRelayNo - '0';
          relayAction(iRelayNo, sAction);
          oNanoESP.sendDataClose(id, getRelayStatus(iRelayNo));
        }
        if (sAction =="ON") {                                 // initialize switch off after max time
          bIsRelayOn = true;
          lRollTime = millis() + MAXMIN;
          Serial.println(String(lRollTime));
        }
      }
    }
    else {                                                    // all other cases => 404 errorpage
      oHttp.sendFromFlash(id, error404, sizeof(error404));    //send website from website.h
      oNanoESP.closeConnection(id);                           //close connection so browser knows website transmission is complete
    }
  }

  if (bIsRelayOn == true && millis() >= lRollTime){           // emergency; because if WIFI connect is loss, you have no chance to switch off => worst case: reservoir is empty
    for (i=8;i>=1;i--) {
      switchRelayOff(i);
    }
    Serial.println("switch off (max time) completed");
    bIsRelayOn = false;
  }
 
}

//------- Start some additional functions ----------------------------------------------------------------------

void relayAction(int iRelayNo, String sAction) {              // switch relay ON or OFF
  if (iRelayNo >= 1 && iRelayNo <= 8  && sAction != "") {
    if (sAction == "ON") {
      switchRelayOn(iRelayNo);
    } else if (sAction == "OFF") {
      switchRelayOff(iRelayNo);
    }
  }
}

void switchPumpOn() {                                         // switch pump relay to ON, if not yet ON
  if (digitalRead(R1) == HIGH) {
    digitalWrite(R1, LOW);
    delay(1000);                                              // one second between pump ON and any other relay ON to build pressure on the valve
  }
}

void switchRelayOn(int iRelayNo) {                            // set one relay ON
  if (iRelayNo == 1) {                                        // 1 = pump relay
    digitalWrite(R1, LOW);
  } else {
    for (i=2;i<=8;i++) {                                      // first switch all "others" OFF to avoid a high current
      if (i!=iRelayNo) {
         switchRelayOff(i);
      }
    }
    switchPumpOn();                                           // pump have to be ON before (if not yet ON)
    if (iRelayNo == 2) {                                      // set only ONE relay to ON
      digitalWrite(R2, LOW);
    } else if (iRelayNo == 3) {
      digitalWrite(R3, LOW);
    } else if (iRelayNo == 4) {
      digitalWrite(R4, LOW);
    } else if (iRelayNo == 5) {
      digitalWrite(R5, LOW);
    } else if (iRelayNo == 6) {
      digitalWrite(R6, LOW);
    } else if (iRelayNo == 7) {
      digitalWrite(R7, LOW);
    } else if (iRelayNo == 8) {
      digitalWrite(R8, LOW);
    }
  }
}

void switchRelayOff(int iRelayNo) {                         // set relay OFF
  if (iRelayNo == 1) {
    digitalWrite(R1, HIGH);
  } else if (iRelayNo == 2) {
    digitalWrite(R2, HIGH);
  } else if (iRelayNo == 3) {
    digitalWrite(R3, HIGH);
  } else if (iRelayNo == 4) {
    digitalWrite(R4, HIGH);
  } else if (iRelayNo == 5) {
    digitalWrite(R5, HIGH);
  } else if (iRelayNo == 6) {
    digitalWrite(R6, HIGH);
  } else if (iRelayNo == 7) {
    digitalWrite(R7, HIGH);
  } else if (iRelayNo == 8) {
    digitalWrite(R8, HIGH);
  }
}

String getRelayStatus(int iRelayNo) {                       // get relay status
  String sRelayStatus = "";
  sRelayStatus = "ON";
  if (iRelayNo == 1) {
    if (digitalRead(R1) == HIGH) {
      sRelayStatus = "OFF";
    }
  } else if (iRelayNo == 2) {
    if (digitalRead(R2) == HIGH) {
      sRelayStatus = "OFF";
    }
  } else if (iRelayNo == 3) {
    if (digitalRead(R3) == HIGH) {
      sRelayStatus = "OFF";
    }
  } else if (iRelayNo == 4) {
    if (digitalRead(R4) == HIGH) {
      sRelayStatus = "OFF";
    }
  } else if (iRelayNo == 5) {
    if (digitalRead(R5) == HIGH) {
      sRelayStatus = "OFF";
    }
  } else if (iRelayNo == 6) {
    if (digitalRead(R6) == HIGH) {
      sRelayStatus = "OFF";
    }
  } else if (iRelayNo == 7) {
    if (digitalRead(R7) == HIGH) {
      sRelayStatus = "OFF";
    }
  } else if (iRelayNo == 8) {
    if (digitalRead(R8) == HIGH) {
      sRelayStatus = "OFF";
    }
  } else {
    sRelayStatus = "--";
  }
  return sRelayStatus;
}

String getRelayStatusAll() {                                // get status of all relais as a common string e.g. ON|OFF|OFF|ON|OFF...
  String sRelayStatus = "";
  for (i=1;i<=8;i++) {
    sRelayStatus += getRelayStatus(i);
    if (i<8)sRelayStatus += "|";
  }
  return sRelayStatus;
}

long getDistance() {                                        // measure of distance
  unsigned long duration;
  digitalWrite(TRIGGER, LOW); 
  delay(5); 
  digitalWrite(TRIGGER, HIGH); 
  delay(10);
  digitalWrite(TRIGGER, LOW);
  duration = pulseIn(ECHO, HIGH); 
  lDistance = (duration/2) * 0.3432;                     //Calculation: division by two (calculation for one way only), multiplied by speed of sound [centimeter/microseconds]*10 => result unit: millimeter

  return (lDistance);
}
